// vi: set ts=4 sw=4 :
// vim: set tw=75 :

// From SDK dlls/h_export.cpp:

/***
*
*	Copyright (c) 1999, 2000 Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
*   Use, distribution, and modification of this source code and/or resulting
*   object code is restricted to non-commercial enhancements to products from
*   Valve LLC.  All other use, distribution, or modification is prohibited
*   without written permission from Valve LLC.
*
****/
/*

===== h_export.cpp ========================================================

  Entity classes exported by Halflife.

*/

#include <extdll.h>

#include <h_export.h>

#include <meta_api.h>		// of course

/*#include <vector>
#include <string>

using namespace std;
*/

#define MSG_BYTE 1
#define MSG_CHAR 2
#define MSG_SHORT 3
#define MSG_LONG 4
#define MSG_ANGLE 5
#define MSG_COORD 6
#define MSG_STRING 7
#define MSG_ENTITY 8
struct data_t
{
	int type;
	int iValue;
	float flValue;
	char sz[256];
	data_t* next;
};
struct msg_t
{
	int dest;
	int type;
	Vector origin;
	edict_t *ed;
	data_t* data;
	//vector<data_t*> data;
};
//vector<msg_t*> msgs;
msg_t *tmp;

// Must provide at least one of these..
static META_FUNCTIONS gMetaFunctionTable;

// Description of plugin
plugin_info_t Plugin_info = {
	META_INTERFACE_VERSION,	// ifvers
	"MESSAGE",	// name
	"1.2",	// version
	"2002/04/07",	// date
	"",	// author
	"",	// url
	"MESSAGE",	// logtag
	PT_ANYTIME,	// (when) loadable
	PT_ANYPAUSE,	// (when) unloadable
};

// Global vars from metamod:
meta_globals_t *gpMetaGlobals;		// metamod globals
gamedll_funcs_t *gpGamedllFuncs;	// gameDLL function tables
mutil_funcs_t *gpMetaUtilFuncs;		// metamod utility functions

// Metamod requesting info about this plugin:
//  ifvers			(given) interface_version metamod is using
//  pPlugInfo		(requested) struct with info about plugin
//  pMetaUtilFuncs	(given) table of utility functions provided by metamod
C_DLLEXPORT int Meta_Query(char *ifvers, plugin_info_t **pPlugInfo,
		mutil_funcs_t *pMetaUtilFuncs) 
{
	if(ifvers);	// to satisfy gcc -Wunused
	// Give metamod our plugin_info struct
	*pPlugInfo=&Plugin_info;
	// Get metamod utility function table.
	gpMetaUtilFuncs=pMetaUtilFuncs;
	return(TRUE);
}

// Metamod attaching plugin to the server.
//  now				(given) current phase, ie during map, during changelevel, or at startup
//  pFunctionTable	(requested) table of function tables this plugin catches
//  pMGlobals		(given) global vars from metamod
//  pGamedllFuncs	(given) copy of function tables from game dll
C_DLLEXPORT int Meta_Attach(PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, 
		meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs) 
{
	if(now);	// to satisfy gcc -Wunused
	if(!pMGlobals) {
		LOG_ERROR(PLID, "Meta_Attach called with null pMGlobals");
		return(FALSE);
	}
	gpMetaGlobals=pMGlobals;
	if(!pFunctionTable) {
		LOG_ERROR(PLID, "Meta_Attach called with null pFunctionTable");
		return(FALSE);
	}
	memcpy(pFunctionTable, &gMetaFunctionTable, sizeof(META_FUNCTIONS));
	gpGamedllFuncs=pGamedllFuncs;
	return(TRUE);
}

// Metamod detaching plugin from the server.
// now		(given) current phase, ie during map, etc
// reason	(given) why detaching (refresh, console unload, forced unload, etc)
C_DLLEXPORT int Meta_Detach(PLUG_LOADTIME now, PL_UNLOAD_REASON reason) {
	if(now && reason);	// to satisfy gcc -Wunused
	return(TRUE);
}


//! Holds engine functionality callbacks
enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;

/*
#define	MSG_BROADCAST		0		// unreliable to all
#define	MSG_ONE				1		// reliable to one (msg_entity)
#define	MSG_ALL				2		// reliable to all
#define	MSG_INIT			3		// write to the init string
#define MSG_PVS				4		// Ents in PVS of org
#define MSG_PAS				5		// Ents in PAS of org
#define MSG_PVS_R			6		// Reliable to PVS
#define MSG_PAS_R			7		// Reliable to PAS
#define MSG_ONE_UNRELIABLE	8		// Send to one client, but don't put in reliable stream, put in unreliable datagram ( could be dropped )
#define	MSG_SPEC			9		// Sends to all spectator proxies
*/

int msg_type=0;
int model_index=0;

char* UTIL_Convert(const char* string)
{
	static char message[1024];
	static int len; len = 0;
	
	if(string){
		while (*string){
			if (*string == '\\')	{
				switch(*(string + 1)){
				case '\\':
					message[len++] = *(string+1);
					message[len++] = *(string+2);
					string+=3;
					continue;
				case 'n':
					message[len++] = '\n';
					string+=2;
					continue;
				case '\'':
					message[len++] = '\"';
					string+=2;
					continue;
				}
			}
			
			message[len++] = *string++;
		}
	}
	
	message[len]='\0';
	return message;
}

void MessageBegin(void) {
	//MESSAGE_BEGIN dest type index x y z
	int argc=CMD_ARGC();
	if(msg_type!=0)return;

	if(argc>=3)
	{
		int msg_dest=atoi(CMD_ARGV(1));
		if(msg_dest<MSG_BROADCAST||MSG_SPEC<msg_dest)return;
		const char *msg=CMD_ARGV(2);
		int index=0;
		edict_t *ed=NULL;
		if(argc>=4)index=atoi(CMD_ARGV(3));
		if(0<index&&index<=gpGlobals->maxClients)
		{
			ed=INDEXENT(index);
		}
		Vector origin=Vector(0,0,0);
		if(argc>=7)
		{
			origin.x=atoi(CMD_ARGV(4));
			origin.y=atoi(CMD_ARGV(5));
			origin.z=atoi(CMD_ARGV(6));
		}
		if(strcmpi(msg,"TE")==0)
		{
			msg_type=SVC_TEMPENTITY;
		}
		else
		{
			msg_type=GET_USER_MSG_ID(PLID, msg, NULL);
			if(msg_type==0)
			{
				msg_type=REG_USER_MSG( msg, -1 );
			}
		}
		if(msg_type!=0)
		{
		//	MESSAGE_BEGIN(msg_dest,msg_type,origin,ed);
			tmp=new msg_t;
			tmp->dest=msg_dest;
			tmp->type=msg_type;
			tmp->origin=origin;
			tmp->ed=ed;
			tmp->data=NULL;

		}
	}
	else
	{
		ALERT( at_console, "usage: MESSAGE_BEGIN msg_dest msg_type clientindex x y z\n" );
	}

}
void MessageEnd(void) {
	if(msg_type!=0)
	{

	//	msgs.push_back(tmp);
		///
		MESSAGE_BEGIN(tmp->dest,tmp->type,tmp->origin,tmp->ed);
	//	unsigned data_size = tmp->data.size();
	//	for ( unsigned i = 0; i < data_size; i++ )
		data_t *newdata=NULL,*next=NULL;
		data_t *data = tmp->data;
		while ( data )
		{
			next = data->next;
			data->next = newdata;
			newdata = data;
			data= next;
		}
		tmp->data = newdata;
		data=tmp->data;

		while(data)
		{
	
			switch(data->type)
			{
			case MSG_BYTE:
				{
					WRITE_BYTE(data->iValue);
				}
				break;
			case MSG_CHAR:
				{
					WRITE_CHAR(data->iValue);
				}
				break;
			case MSG_SHORT:
				{
					WRITE_SHORT(data->iValue);
				}
				break;
			case MSG_LONG:
				{
					WRITE_LONG(data->iValue);
				}
				break;
			case MSG_ANGLE:
				{
					WRITE_ANGLE(data->flValue);
				}
				break;
			case MSG_COORD:
				{
					WRITE_COORD(data->flValue);
				}
				break;
			case MSG_STRING:
				{
					WRITE_STRING(data->sz);
				}
				break;
			case MSG_ENTITY:
				{
					WRITE_ENTITY(data->iValue);
				}
				break;
			}
			data=data->next;
		}
		MESSAGE_END();

		tmp->dest=0;
		tmp->type=0;
		tmp->origin=Vector(0,0,0);
		tmp->ed=0;
		data=tmp->data;
		data_t *rem=NULL;
		while(data)
		{
			rem=data->next;
			delete data;
			data=rem;
		}
		tmp->data=NULL;
		delete tmp;
		tmp=0;
		msg_type=0;
		model_index=0;
	}
}

void WriteByte(void) {

	if(msg_type!=0&&CMD_ARGC()>1)
	{
		int iValue=atoi(CMD_ARGV(1));
	//	WRITE_BYTE(iValue);
		data_t *data=new data_t;
		data->type=MSG_BYTE;
		data->iValue=iValue;
		data->next=tmp->data;
		tmp->data=data;
	}
}
void WriteChar(void) {
	if(msg_type!=0&&CMD_ARGC()>1)
	{
		int iValue=atoi(CMD_ARGV(1));
	//	WRITE_CHAR(iValue);
		data_t *data=new data_t;
		data->type=MSG_CHAR;
		data->iValue=iValue;
		data->next=tmp->data;
		tmp->data=data;
	}
}
void WriteShort(void) {
	if(msg_type!=0&&CMD_ARGC()>1)
	{
		int iValue;
		if(strcmpi(CMD_ARGV(1),"model")==0)
		{
			iValue=model_index;
		}
		else
		{
			iValue=atoi(CMD_ARGV(1));
		}	
		//WRITE_SHORT(iValue);
		data_t *data=new data_t;
		data->type=MSG_SHORT;
		data->iValue=iValue;
		data->next=tmp->data;
		tmp->data=data;
	}
}
void WriteLong(void) {
	if(msg_type!=0&&CMD_ARGC()>1)
	{
		int iValue=atoi(CMD_ARGV(1));
	//	WRITE_LONG(iValue);
		data_t *data=new data_t;
		data->type=MSG_LONG;
		data->iValue=iValue;
		data->next=tmp->data;
		tmp->data=data;
	}
}
void WriteAngle(void) {
	if(msg_type!=0&&CMD_ARGC()>1)
	{
		float flValue=atof(CMD_ARGV(1));
	//	WRITE_ANGLE(flValue);
		data_t *data=new data_t;
		data->type=MSG_ANGLE;
		data->flValue=flValue;
		data->next=tmp->data;
		tmp->data=data;
	}
}
void WriteCoord(void) {
	if(msg_type!=0&&CMD_ARGC()>1)
	{
		float flValue=atof(CMD_ARGV(1));
		//WRITE_COORD(flValue);
		data_t *data=new data_t;
		data->type=MSG_COORD;
		data->flValue=flValue;
		data->next=tmp->data;
		tmp->data=data;
	}
}
void WriteString(void) {
	if(msg_type!=0&&CMD_ARGC()>1)
	{

	//	WRITE_STRING(sz);
		data_t *data=new data_t;
		data->type=MSG_STRING;
		strncpy(data->sz,UTIL_Convert(CMD_ARGS()),255);
		data->next=tmp->data;
		tmp->data=data;
	}
}
void WriteEntity(void) {
	if(msg_type!=0&&CMD_ARGC()>1)
	{
		int iValue=atoi(CMD_ARGV(1));
	//	WRITE_ENTITY(iValue);
		data_t *data=new data_t;
		data->type=MSG_ENTITY;
		data->iValue=iValue;
		data->next=tmp->data;
		tmp->data=data;
	}
}
void ModelIndex(void) {
	if(CMD_ARGC()>1)
	{
		model_index=MODEL_INDEX(CMD_ARGV(1));
	}
}

// Receive engine function table from engine.
// This appears to be the _first_ DLL routine called by the engine, so we
// do some setup operations here.
void WINAPI GiveFnptrsToDll( enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals )
{
	memcpy(&g_engfuncs, pengfuncsFromEngine, sizeof(enginefuncs_t));
	gpGlobals = pGlobals;
	REG_SVR_COMMAND("MESSAGE_BEGIN",MessageBegin);
	REG_SVR_COMMAND("MESSAGE_END",MessageEnd);
	REG_SVR_COMMAND("WRITE_BYTE",WriteByte);
	REG_SVR_COMMAND("WRITE_CHAR",WriteChar);
	REG_SVR_COMMAND("WRITE_SHORT",WriteShort);
	REG_SVR_COMMAND("WRITE_LONG",WriteLong);
	REG_SVR_COMMAND("WRITE_ANGLE",WriteAngle);
	REG_SVR_COMMAND("WRITE_COORD",WriteCoord);
	REG_SVR_COMMAND("WRITE_STRING",WriteString);
	REG_SVR_COMMAND("WRITE_ENTITY",WriteEntity);
	REG_SVR_COMMAND("SET_MODEL",ModelIndex);
}
